#!/usr/local/bin/python
#
# adzapper
# Copyright 1999 Adam Feuer
# Adam Feuer <adamf@pobox.com>
#
# This file is part of adzapper.
#
# adzapper is free software; you can redistribute it and/or modify
# it under the terms of the Zope Public License 1.0 as published by
# Digital Creations.
#
# adzapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Zope Public License for more details.
#
# You should have received a copy of the Zope Public License 1.0
# along with adzapper; see the file COPYING.
#
#  "This product includes software developed by Digital Creations
#  for use in the Z Object Publishing Environment
#  (http://www.zope.org/)."
#
#
# adzapper is an http proxy that filters out ads
#		
#
"""adzapper startup file
Usage: %(program)s [Options] [<port>]
Options:

-h
    Output this text.

-d <debug level>
    Turn debugging on. 0 is off; higher numbers produce greater detail. 

-p
    Turn profiling on

-v
    Display version number

-n <nameserver ip address>
    Use nameserver at this ip address

-z <zaplet directory>
    Path to zaplet directory. Default is %(here)s/zaplets. If the directory
    name does not start with a '/' it is interpreted relative to the current
    working directory.

-f <zaplet config file>
    Name of zaplet configuration file. Default is
    %(here)s/adzapper.conf. If the file name does not start with a '/'
    it is interpreted relative the current working directory.

<port>    
    Port to run on. Default is port 51966.

"""

RCS_ID =  '$Id: adzapper.py,v 1.6 1999/10/10 03:23:45 adamf Exp $'

# global variables
MAJOR_VERSION = 0
MINOR_VERSION = 1
SUB_VERSION   = 30     # ;-)  

# zaplet config stuff
ZAPLET_DIR = 'zaplets'
ADZAPPER_CONFIG_FILE = 'adzapper.conf'

# standard modules
import string
import getopt
import os
import sys

# holder object for adzapper_engine and zaplet_engine
class combined_engine:
    def __init__(self):
        self.adzapper = None
        self.zaplet = None


if __name__ == '__main__':
    # fix the path
    program=sys.argv[0]
    here=os.path.join(os.getcwd(), os.path.split(program)[0])
    sys.path=[os.path.join(here,'src'),os.path.join(here,'src','medusa'),here] + filter(None, sys.path)

    # import modules.
    import debug

    # fix asyncore signal handling
    import asyncore
    import asyncore_fixes
    asyncore.poll = asyncore_fixes.a_poll

    # fix asyncore logging 
    asyncore.dispatcher.log_info = debug.log_info
    asyncore.dispatcher.log = debug.log

    # fix http_server.http_channel writable predicate
    import http_server
    import http_server_fixes
    http_server.http_channel.writable = http_server_fixes.writable
    delattr(http_server.http_channel, 'writable_for_proxy')
    http_server.http_channel.handle_close = http_server_fixes.handle_close


    import resolver
    
    # adzapper and zaplet modules
    import platform_dependent
    import http_proxy
    import adzapper_engine
    import zaplet_engine

    # debugging defaults to off
    debug.debuglevel = 0

    # profiling defaults to off
    profiling_on = 0

    # we don't have the nameserver IP address yet
    nameserver = None

    # set the defaults for zaplet directory and config file
    zaplet_dir = ZAPLET_DIR
    adzapper_config_file = ADZAPPER_CONFIG_FILE

    program_name = sys.argv[0]
    args = sys.argv[1:]
    debug.debug (1,("args: ",  args))

    try:
        optlist, arglist = getopt.getopt(args, 'hpvd:n:z:f:')
    except getopt.error, error_message:
        print "error message: %s" % error_message
        print
        print __doc__ % vars()
        raise SystemExit

    debug.debug(1,('args: ' , args))
    debug.debug(1,('optlist' , optlist))
    debug.debug(1,('arglist: ' , arglist))

    if arglist == []:
        port = 51966
    else:
        debug.debug(1, 'port: %s ' % arglist[0])
        port = string.atoi(arglist[0])

    if '-h' in args:
        print __doc__ % vars()
        raise SystemExit

    if '-v' in args:
        VERSION_STRING = '%d.%d.%d' % (MAJOR_VERSION,MINOR_VERSION,SUB_VERSION)
        print '%s version: %s' % (program_name,VERSION_STRING)
        raise SystemExit

    if '-p' in args:
        profiling_on = 1

    if '-d' in args:
        for list_item in optlist:
            if list_item[0] == '-d':
                debug.debuglevel = string.atoi(list_item[1])

    if '-n' in args:
        for list_item in optlist:
            if list_item[0] == '-n':
                nameserver = list_item[1]

    if '-z' in args:
	# note: if more then one '-z' is specified,
	# the last one on the command line is used.
        for list_item in optlist:
            if list_item[0] == '-z':
                zaplet_dir = list_item[1]
                
    if '-f' in args:
        for list_item in optlist:
            if list_item[0] == '-f':
                adzapper_config_file = list_item[1]

    debug.debug(1,'Debugging output is turned on at level %d' % debug.debuglevel)

    if (nameserver == None):
	# get the nameserver IP address for resolver
        nameserver = platform_dependent.getNameServer()
        if (nameserver == None):
	    # nameserver defaults to localhost
            nameserver = '127.0.0.1'
    debug.debug(1,"nameserver: %s" % nameserver)

    # make the persistent resolver
    #  and set it to use our log_info
    debug.debug(2,'Using nameserver %s' % nameserver)
    resolver = resolver.caching_resolver (nameserver)

    # make the holder object for the combined adzapper and zaplet engines
    ce = combined_engine()

    # initialize adzapper engine - for adzapper defaults, ACLs, etc.
    ce.adzapper = adzapper_engine.adzapper_engine(adzapper_config_file)

    # initialize zaplet engine - for blocking URLs
    ce.zaplet = zaplet_engine.zaplet_engine(zaplet_dir)

    # install signal handler to catch the signal SIGHUP 
    # (only do this for Unix platforms)
    platform_dependent.setSighupHandler(ce.zaplet.__init__,
					zaplet_dir)

    # make the handler- need to give it the resolver and the combined_engine
    ph = http_proxy.http_proxy_handler(resolver,ce)
    hs = http_server.http_server ('', port)

    hs.install_handler (ph)

    if profiling_on:
        def profile_loop ():
            try:
                asyncore.loop()
            except KeyboardInterrupt:
                pass
            import profile
            profile.run ('profile_loop()', 'profile.out')
    else:
        asyncore.loop()

