#
# This file is part of Zaplet
# Copyright 1999 Adam Feuer <adamf@pobox.com>
#
# Zaplet is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Zaplet is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Zaplet; see the file COPYING-Zaplet. If not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#

# zaplet classes
# for use with adzapper filtering proxy

RCS_ID =  '$Id: zaplet_engine.py,v 1.6 1999/10/03 01:41:18 adamf Exp $'

import os
import re
import string
import os
import sys
import urlparse
import socket

# adzapper modules
import debug

numeric_host = re.compile ('[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+')
numeric_network = re.compile ('[0-9]+\.([0-9]+\.)?([0-9]+\.)?([[0-9]+\*])?')


def quoted_string(arg):
    """returns true if string is enclosed in quotes"""
    l = len(arg)
    if l > 1:
        if arg[0] == "\"" and arg[l-1] == "\"":
            return 1
        elif arg[0] == "\'" and arg[l-1] == "\'":
            return 1
        return 0


def regex_string(arg):
    """returns true of string is enclosed in '/'"""
    l = len(arg)
    if l > 1:
        if arg[0] == '/' and arg[l-1] == '/':
            return 1
    return 0


def trim(arg):
    """cut off the first and last characters of the string"""
    l = len(arg)
    if l > 2:
        return arg[1:l-1]
    else:
        return ""



class zaplet:
    """data structure object to hold zaplet"""
    def __init__(self):
        # zaplet version defaults to 0.1
        self.majorversion = None
        self.minorversion = None
        self.host = None
        self.allow = ''
        self.block = ''
        self.lines = []




class zaplet_engine:
    """Zaplet engine class: this does all the work--
    reads in zaplets, and determines whether URLs should be blocked."""

    def __init__(self, zDir):
        """initialize the zaplet_engine object
        build the {site : block_regex} dictionary"""
        
        self.zDir = zDir

        # dictionary of (site, zaplet object instance) pairs
        self.zaplet_list = { }

        self.readZaplets(self.zDir)



    def readZaplets(self,zDir):
        """read Zaplets from the repositories
        note: zDir must have a trailing '/' """
        debug.debug(1,"Initializing zaplet engine...")
        cpairs = []  # list of (site, conf file) pairs

        # make the list of zaplets

        if (zDir[0] == '/'):
            # it is an absolute path
            pass
        else:
            # not absolute path to zaplet directory-- add cwd to it
            self.zDir = os.path.join(os.getcwd(),zDir)
        debug.debug(1, "zaplet_engine, zaplet dir: %s" % self.zDir)

        zdirlist = os.listdir(self.zDir)
        debug.debug(1, "zaplet_engine, zdirlist:")
        debug.debug(1, zdirlist)

        zapfile =  re.compile('^.*\.zap')
        dotfile =  re.compile('^\.')
        conffile = re.compile('.*\.conf$')
        zlist = []
        for file in zdirlist:
            if os.path.isfile(os.path.join(self.zDir,file)):
                if zapfile.search(file) and not dotfile.search(file) and not conffile.search(file):
                    zlist.append(file)
        debug.debug(1,"processed source directory: ")
        debug.debug(1,zlist)

        # now go thru list of zaplets, read them in and compile them.
        for zfile in zlist:
            debug.debug(1,"site: " + zfile + "        zaplet file: " + zfile)
            zl = self.get_zaplet(os.path.join(self.zDir,zfile))

            # if we got a list of zaplets back...
            if len(zl) > 0:
                # add each of the zaplets to the dictionary
                for zlitem in zl:
                    self.zaplet_list[zlitem.host] = zlitem

        debug.debug(1,self.zaplet_list.keys())



    def get_zaplet(self,zapletfile):
        """make the list of lists of zaplet lines
        given the name of a zapletfile, read it in, split any zaplets into separate lists of lines,
        and call make_zaplet() on each list of lines.
        note: zapletfile must be an absolute path to the zaplet file"""
        
        zaplist = []

        try:
            debug.debug(1," opening zaplet : " + zapletfile)
            cf = open(zapletfile, "r")
        except:
            sys.stderr.write("Cannot open zaplet " + zapletfile + "\n")
            return None

        lines = cf.readlines()
        cf.close()

        in_zaplet = 0
        for line in lines:
            if in_zaplet:
                zlines.append(line)
                if re.search('^\s*</zaplet>',line):
                    in_zaplet = 0
                    zaplist.append(zlines)
                    debug.debug(1,zlines)
            else: # not in_zaplet
                if re.search('^\s*<zaplet>',line):
                    in_zaplet = 1
                    zlines = []
                    zlines.append(line)

        zaplet_list = []
        for zlitem in zaplist:
            tmp = self.make_zaplet(zlitem)
            if tmp != None:
                zaplet_list.append(tmp)
            else:
                debug.debug(1,"%s: zaplet not used." % zapletfile)

        return zaplet_list



    def make_zaplet(self,zapletlines):
        """make a new zaplet
        given the list of lines that make a zaplet, parse the lines and return a zaplet object
        complete with compiled regexes

        returns None if zaplet is invalid."""
        
        z = zaplet()

        block_re_str = '' # the block regex that will be built for this zaplet
        allow_re_str = '' # the allow regex that will be built for this zaplet

        z.lines = zapletlines

        in_zaplet = 0
        for line in z.lines:
            if in_zaplet:
                line = string.strip(line)
                if (len(line) == 0) or re.search("\s*#", line):
                    continue
                if string.find(line, '#'):
                    ltmp = string.split(line, '#')
                    line = ltmp[0]
                tokens = re.split('\s+', line)
                command = string.lower(tokens[0])
                if (command == 'version'):
                    vstr = tokens[1]
                    vlist = string.split(vstr,'.')
                    z.majorversion = string.atoi(vlist[0])
                    z.minorversion = string.atoi(vlist[1])
                    debug.debug(1,'zaplet version: %i.%i' % (z.majorversion, z.minorversion))
                elif (command == 'host'):
                    z.host = string.lower(tokens[1])
                    debug.debug(1,'zaplet host: %s' % z.host)
                elif (command == 'block'):
                    if len(tokens) > 1:
                        arg = tokens[1]
                    else:
                        arg = ""
                    if regex_string(arg):  
                        # regular expression - don't escape
                        # trim slashes and add to regex
                        debug.debug (2,"allow regular expression: %s" % arg)
                        re_add = trim(arg)
                        debug.debug(2,"block trimmed regex: %s" % re_add)
                        block_re_str =  block_re_str + '|' + re_add 
                    elif quoted_string(arg):
                        # literal string - trim quotes & escape
                        debug.debug (2,"block quoted string: %s" % arg)
                        re_add = trim(arg)
                        debug.debug(2, "block trimmed string: %s" % re_add)
                        re_add = re.escape(re_add)
                        block_re_str = block_re_str + '|' + re_add
                    else:   # literal string - need to escape it.
                        re_add = re.escape(arg)
                        block_re_str = block_re_str + '|' + re_add
                        debug.debug(2, "block literal string: %s" % re_add)
                elif (command == 'allow'):
                    if len(tokens) > 1:
                        debug.debug(2, "tokens[1]: %s" % tokens[1])
                        arg = tokens[1]
                    else:
                        arg = ""
                    if regex_string(arg):  
                        # regular expression - don't escape
                        # trim slashes and add to regex
                        debug.debug (2,"allow regular expression: %s" % arg)
                        re_add = trim(arg)
                        debug.debug(2,"allow trimmed regex: %s " % re_add)
                        allow_re_str =  allow_re_str + '|' + re_add 
                    elif quoted_string(arg):
                        # literal string - trim quotes & escape
                        debug.debug (2,"allow quoted string: %s" % arg)
                        re_add = trim(arg)
                        debug.debug(2, "allow trimmed string: %s" % re_add)
                        re_add = re.escape(re_add)
                        allow_re_str = allow_re_str + '|' + re_add
                    else:   # literal string - need to escape it.
                        re_add = re.escape(arg)
                        allow_re_str = allow_re_str + '|' + re_add
                        debug.debug(2, "allow literal string: %s" % re_add)

                elif (command == '</zaplet>'):
                    in_zaplet = 0
                    continue
                else:
                    debug.debug(1,"unknown command: " + command)
                    pass
            else: # not in_zaplet
                if re.search('^<zaplet>',line):
                    in_zaplet = 1
                    continue

        # if there is no version statement, don't use the zaplet--
        #   version statements are now required!
        if z.majorversion != None and z.minorversion != None:
            debug.debug(1,"z.version: %d.%d " % (z.majorversion,z.minorversion))
        else:
            debug.debug(1,"zaplet not used; no version statement!")
            return None

        # if there is no host statement, don't use the zaplet--
        #   host statements are now required
        if z.host != None:
            debug.debug(1,"z.host: %s " % z.host)
        else:
            debug.debug(1,"zaplet not used; no host statement!")
            return None


        # chop off initial |
        allow_re_str = allow_re_str[1:]
        block_re_str = block_re_str[1:]

        debug.debug(1,"allow_re_str: %s" % allow_re_str)
        debug.debug(1,"block_re_str: %s" % block_re_str)

        # compile regexes
        if allow_re_str == "":
            allow_re_str = "^$"
        z.allow = allow_re_str
        z.allow_re = re.compile(z.allow)

        z.block = block_re_str
        z.block_re = re.compile(z.block)

        return z



    def checkNamedHost(self,host,url):
        """check to see whether a named host is blocked -
        uses zaplet database, searchs from most specific to least specific
        (whole dns name www.foo.bar.com, then foo.bar.com, then bar.com, then .com)
        returns 1 if url should be blocked, 0 if not."""

        srv_fields = string.split(host,'.')
        #debug.debug(1,srv_fields)
        while (srv_fields != []):
            tmps = string.join(srv_fields,'.');
            if (self.zaplet_list.has_key(tmps)):
                # we are on the site-- check for allows first
                if (self.zaplet_list[tmps].allow_re.search(url)):
                    # it's specifically allowed, so let it thru
                    debug.debug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                    return 0
                # ok, it's not specifically allowed, now check for blocks
                elif (self.zaplet_list[tmps].block_re.search(url)):
                    debug.debug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                    return 1
                # it's not specifically blocked, so let it thru.
                else:
                    # return-- we don't want to continue searching
                    return 0
            del srv_fields[0]
        return 0




    def checkNumericHost(self,host,url):
        """check to see whether a numeric host is blocked -
        uses zaplet database, searches from most specific to least specific
        (whole ip number a.b.c.d, then network a.b.c, then net a.b, and then net a)
        returns 1 if url should be blocked, 0 if not."""
        
        srv_fields = string.split(host,'.')
        debug.debug(1,"numeric host ip: %s" % srv_fields)
        lastfield = len(srv_fields)
        debug.debug(1,"numeric host ip fields: %d" % lastfield)
        for i in range(0,lastfield):
            tmps = string.join(srv_fields[0:lastfield-i],'.');
            debug.debug(1,"numeric host- iteration ip: %s" % tmps)
            if (self.zaplet_list.has_key(tmps)):
                # we are on the site-- check for allows first
                if (self.zaplet_list[tmps].allow_re.search(url)):
                    # it's specifically allowed, so let it thru
                    debug.debug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                    return 0
                # ok, it's not specifically allowed, now check for blocks
                elif (self.zaplet_list[tmps].block_re.search(url)):
                    debug.debug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                    return 1
                # it's not specifically blocked, so let it thru.
                else:
                    # return-- we don't want to continue searching
                    return 0
                # default zaplet

        tmps = 'default-numeric'
        debug.debug(1,"checking default-numeric...")
        if (self.zaplet_list.has_key(tmps)):
            if (self.zaplet_list[tmps].allow_re.search(url)):
                # it's specifically allowed, so let it thru
                debug.debug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                return 0
            # ok, it's not specifically allowed, now check for blocks
            elif (self.zaplet_list[tmps].block_re.search(url)):
                debug.debug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                return 1
        return 0


 

    def urlblocker(self, url):
        """determine whether a URL should be zapped or not
        returns 1 if it should be blocked, 0 otherwise"""
        debug.debug(1,"zap: in zap.")
        url_tuple = urlparse.urlparse(url)
        host_port = url_tuple[1]
        debug.debug(1,"zaplet.urlblocker- host_port: %s" % host_port)

        # deal with urls of the form "http://user:password@host.domain.com/"
        #   as per RFC1738
        # thanks to "Constantinos A. Kotsokalis" <C.Kotsokalis@ece.ntua.gr>
        rem_host = ''
        rem_port = 80
        if (string.find(host_port, ':') != -1):
            occ = string.count(host_port, ':')
            atocc = string.count(host_port, '@')
            if occ == 1 and atocc == 0:
                rem_host, rem_port = string.split(host_port, ':')
                rem_port = string.atoi(rem_port)
            elif occ == 1 and atocc == 1:
                rem_host, rem_port = string.split(host_port, '@')[1], 80
            elif occ == 2:
                rem_host, rem_port = string.split(string.split(host_port, '@')[1], ':')
                rem_port = string.atoi(rem_port)
        else:
            rem_host, rem_port = host_port, 80

        rem_host=string.lower(rem_host)

        # does the remote host match a zaplet in the database?
        if numeric_host.match(rem_host) == None:
            # has DNS name
            debug.debug(1, "named host.")
            if self.checkNamedHost(rem_host,url) == 1:
                # yes
                return 1
        else:
            # is a numeric host
            debug.debug(1, "numeric host.")
            if self.checkNumericHost(rem_host,url) == 1:
                return 1

        # default zaplet
        debug.debug(1,"checking default...")
        tmps = 'default'
        if (self.zaplet_list.has_key(tmps)):
            debug.debug(2,"zap: has default zaplet...")
            if (self.zaplet_list[tmps].allow_re.search(url)):
                # it's specifically allowed, so let it thru
                debug.debug(2,"zap: allow- matched pattern- site: " + tmps + "    url: " + url)
                return 0
            # ok, it's not specifically allowed, now check for blocks
            elif (self.zaplet_list[tmps].block_re.search(url)):
                debug.debug(2,"zap: block matched pattern- site: " + tmps + "    url: " + url)
                return 1
        debug.debug(2,"zap: no allows and no blocks. (%s)" % url)
        return 0


